
	list p=18f4420
	#include p18f4420.inc
	#include defs.inc
	#include equ.inc


Startup		code	0x0000
	goto	Start
	nop

HighISR		code	0x0008
	btfsc	INTCON, INT0IF
	rcall	KbdInt			; INT0 -> dato da tastiera
	btfsc	INTCON3, INT2IF
	reset					; INT2 -> reset
	bcf		INTCON, INT0IF	; resetta flag INT0
	retfie	FAST

LowISR		code	0x0018
	btfss	INTCON3, INT1IF
	retfie					; INT non generato da INT1 (linea DRQ)
	bcf		INTCON3, INT1IF
	bsf		KFRFlag, KFRFULL	; segnala registri da leggere
	retfie
	

	code

KbdInt:
	rcall	RxKeyb			; riceve byte da tastiera
	SCLOCKL					; disabilita TX da tastiera
	btfsc	KStatus, KRESET	; test reset
	reset					; ricevuto reset in -> reset
	btfsc	KStatus, KTIMEO	; test timeout
	return					; scarta dato ricevuto se timeout
	btfss	KStatus, KBRX	; test ricezione full
	return					; ricezione incompleta -> scarta dato
	ldw		KStatus			; test errore ricezione
	andlw	KBRXERR
	bz		KbdInt10		; dato ricevuto OK
	movlw	KBD_RESEND		; richiesta di re-invio alla tastiera
	rcall	TxKeyb
	btfsc	KStatus, KRESET	; test reset
	reset					; ricevuto reset in -> reset
	return					; ignora altri errori
	
KbdInt10:
	movlw	0x00			; scarta codice 00 (overrun)
	subwf	KBuf, W
	bz		KbdInt99
	movlw	0xFF			; scarta codice FF (overrun/error)
	subwf	KBuf, W
	bz		KbdInt99
	movlw	KBD_RESEND		; scarta richiesta resend
	subwf	KBuf, W
	bz		KbdInt99
	movlw	KBD_ACK			; scarta ACK
	subwf	KBuf, W
	bz		KbdInt99
	movlw	KBD_BAT			; scarta BAT
	subwf	KBuf, W
	bnz		KbdInt11
	bsf		KbdFlag, KBBAT
	bsf		KbdFlag, KBBAT2
	bcf		KbdFlag2, KBNOBAT
	bra		KbdInt99

KbdInt11:
	movlw	KBD_NOBAT			; scarta BAT failure
	subwf	KBuf, W
	bnz		KbdInt12
	bcf		KbdFlag, KBBAT
	bcf		KbdFlag, KBBAT2
	bsf		KbdFlag2, KBNOBAT

KbdInt99:
	SCLOCKH					; abilita TX da tastiera
	return

KbdInt12:						; inserisce byte in coda
	movlw	(RXBUFLEN - 9)		; lascia un limite di guardia di 8 bytes
	subwf	RxCnt,W				; RxCnt - (RXBUFLEN - 9))
	btfsc	STATUS, C			; C = 0 -> OK - spazio disponibile in coda
	bsf		RxFlag, RXOVR		; segnala condizione overrun
	ldw		KBuf				; memorizza byte in coda
	movwf	POSTINC2
	incf	RxCnt, f			; incrementa contatore bytes in coda
	movlw	LOW(RxBufEnd)		; coda circolare - test puntatore
	subwf	FSR2L, W			; FSR2L - LOW(RxBufEnd)
	movlw	HIGH(RxBufEnd)		; limite superiore H
	subwfb	FSR2H, W			; FSR2H - HIGH(RxBufEnd)
	btfsc	STATUS, C			; C = 0 se FSR2 < RxBufEnd
	lfsr	FSR2, RxBuffer		; reimposta puntatore a inizio buffer

KbdInt20:
	btfss	RxFlag, RXOVR	; se in condizioni di overrun disabilita TX
	bra		KbdInt99		; riabilita trasmissione da tastiera
	return					; esce con CLK OUT LOW 

RxKeyb:
	clrf	KStatus			; azzera status
	movlw	D'8'			; contatore 8 bit
	movwf	KCnt
	movlw	D'1'			; parity bit (odd)
	movwf	KPar
	movlw	TMR0P4			; TMR0 stop, prescaler = 4
	movwf	T0CON			; modo 16 bit	
	bcf		INTCON, TMR0IF	; azzera flag timeout
	movlw	HIGH(T20MS)		; imposta timeout
	movwf	TMR0H			; caricare prima TMR0H
	movwf	T0CntH
	movlw	LOW(T20MS)
	movwf	TMR0L			; carica TMR0L/H
	movwf	T0CntL
	bsf		T0CON, TMR0ON	; start Timer 0

RxKeyb1:					; attesa CLK IN -> LOW
	rcall	GetLine
	bmi		RxKeyb1			; CLK IN = HIGH - loop

RxKeyb2:					; loop attesa CLK IN -> HIGH
	rcall	GetLine
	bpl		RxKeyb2			; CLK IN = LOW - loop
	
RxKeyb3:					; loop attesa CLK IN -> LOW
	rcall	GetLine
	bmi		RxKeyb3			; CLK IN = HIGH - loop
	bnc		RxKeyb4			; no parity se DATA IN = LOW
	incf	KPar, f			; incrementa parity 
	bsf		STATUS, C		; forza CARRY = 1 (DATA IN = HIGH)

RxKeyb4:					; aggiorna KBuf
	rrcf	KBuf, f			; KBuf<7> = CARRY
	decfsz	KCnt, f			; decrementa contatore 8 bit
	bra		RxKeyb2			; loop 8 bit

RxKeyb5:					; attesa CLK IN -> HIGH
	rcall	GetLine
	bpl		RxKeyb5			; CLK IN = LOW - loop

RxKeyb6:					; attesa CLK IN -> LOW
	rcall	GetLine
	bmi		RxKeyb6			; CLK IN = HIGH - loop
	rlcf	KCnt, W			; C = parity bit -> W<0> = parity bit
	xorwf	KPar, f			; confronto parity ricevuta con parity calcolata
	rrcf	KPar, W			; C = confronto parity
	bnc		RxKeyb7			; C = 0 -> parity OK
	bsf		KStatus, KPARY	; imposta errore parity

RxKeyb7:					; attesa CLK IN -> HIGH
	rcall	GetLine
	bpl		RxKeyb7			; CLK IN = LOW - loop

RxKeyb8:					; attesa CLK IN -> LOW
	rcall	GetLine
	bmi		RxKeyb8			; CLK IN = HIGH - loop
	bc		RxKeyb9			; C = 1 -> stop bit OK
	bsf		KStatus, KSTOP	; segnala errore bit stop errato

RxKeyb9:					; attesa CLK IN -> HIGH - fine trasmissione
	rcall	GetLine
	bpl		RxKeyb9			; CLK IN = LOW - loop
	bsf		KStatus, KBRX	; segnala ricezione completata
	bcf		T0CON, TMR0ON	; stop Timer 0
	bcf		INTCON, TMR0IF	; azzera flag timeout
	return
 	
GetLine:
	btfsc	INTCON, TMR0IF	; test timeout
	bra		GetLine2		; timeout	
	ldw		PORTB
	andlw	RBMASK			; maschera bit utili
	movwf	TmpB			; salva temporaneo
	ldw		PORTB
	andlw	RBMASK			; maschera bit utili
	cpfseq	TmpB			; confronta con lettura precedente
	bra		GetLine			; attesa linee stabili o timeout
	btfss	TmpB, RESIN		; test asserzione reset da host
	bra		GetLine1		; reset da host
	bcf		STATUS, N		; N = 0 se CLK IN = LOW
	btfsc	TmpB, KCLKIN
	bsf		STATUS, N		; N = 1 se CLK IN = HIGH
	bcf		STATUS, C		; C = 0 se DATA IN = LOW
	btfsc	TmpB, KDATAIN
	bsf		STATUS, C		; C = 1 se DATA IN = HIGH
	return					; C = DATA IN, N = CLK IN

GetLine1:
	bsf		KStatus, KRESET	; segnala reset
	pop						; scarta indirizzo di ritorno
	return					; uscita

GetLine2:
	bsf		KStatus, KTIMEO	; segnala timeout
	bcf		T0CON, TMR0ON	; stop Timer 0
	bcf		INTCON, TMR0IF	; azzera flag timeout
	pop						; scarta indirizzo di ritorno
	return					; uscita


TxKeyb:
	movwf	KBuf			; salva dato da trasmettere
	clrf	KStatus			; azzera status
	movlw	TMR0P0			; TMR0 stop, no prescaler, modo 16 bit
	movwf	T0CON	
	bcf		INTCON, TMR0IF	; azzera flag timeout
	movlw	HIGH(T100US)	; imposta timeout 100us
	movwf	TMR0H			; caricare prima TMR0H
	movlw	LOW(T100US)
	movwf	TMR0L			; carica TMR0L/H
	movlw	D'8'			; contatore 8 bit
	movwf	KCnt
	movlw	D'1'			; parity bit (odd)
	movwf	KPar
	SCLOCKL					; CLK OUT -> LOW - request to send	
	bsf		T0CON, TMR0ON	; start Timer 0

TxKeyb1:					; ritardo 100us
	btfss	INTCON, TMR0IF
	bra		TxKeyb1			; loop ritardo
	SDATAL					; DATA OUT -> LOW (start bit)
	bcf		T0CON, TMR0ON	; stop Timer 0
	bcf		INTCON, TMR0IF	; azzera flag timeout
	movlw	TMR0P4			; TMR0 stop, prescaler = 4
	movwf	T0CON			; modo 16 bit	
	movlw	HIGH(T20MS)		; imposta timeout
	movwf	TMR0H			; caricare prima TMR0H
	movlw	LOW(T20MS)
	movwf	TMR0L			; carica TMR0L/H
	bsf		T0CON, TMR0ON	; start Timer 0
	SCLOCKH					; CLK OUT -> HIGH (rilascia clock)

TxKeyb2:					; attesa CLK IN -> LOW
	rcall	GetLine
	bmi		TxKeyb2			; CLK IN = HIGH -> loop

TxKeyb3:					; loop 8 bit
	rrcf	KBuf			; C = KBuf<0> - bit da inviare
	bnc		TxKeyb4			; C = 0 -> DATA OUT = LOW
	incf	KPar, f			; C = 1 -> aggiorna parity
	SDATAH					; DATA OUT = HIGH
	bra		TxKeyb5

TxKeyb4:
	SDATAL					; DATA OUT = LOW

TxKeyb5:					; invia bit
	rcall	GetLine			; attende CLK IN -> HIGH
	bpl		TxKeyb5			; CLK IN = LOW -> loop

TxKeyb6:
	rcall	GetLine			; attende CLK IN -> LOW
	bmi		TxKeyb6			; CLK IN = HIGH -> loop
	decfsz	KCnt, f			; decrementa contatore 8 bit
	bra		TxKeyb3			; loop 8 bit
	rrcf	KPar, f			; C = KPar<0> -> parity bit
	bnc		TxKeyb7			; C = 0 -> DATA OUT = LOW
	SDATAH					; DATA OUT = HIGH
	bra		TxKeyb8

TxKeyb7:
	SDATAL					; DATA OUT = LOW

TxKeyb8:
	rcall	GetLine			; attende CLK IN -> HIGH
	bpl		TxKeyb8			; CLK IN = LOW -> loop

TxKeyb9:
	rcall	GetLine			; attende CLK IN -> LOW
	bmi		TxKeyb9			; CLK IN = HIGH -> loop
	SDATAH					; DATA OUT -> HIGH - stop bit

TxKeyb10:
	rcall	GetLine			; attende CLK IN -> HIGH
	bpl		TxKeyb10		; CLK IN = LOW -> loop
				
TxKeyb11:
	rcall	GetLine			; attende CLK IN -> LOW
	bmi		TxKeyb11		; CLK IN = HIGH -> loop
	bnc		TxKeyb12		; C = 0 -> DATA IN = LOW - ack bit ok
	bsf		KStatus, KBACK	; segnala errore ack bit

TxKeyb12:
	rcall	GetLine			; attende CLK IN -> HIGH
	bpl		TxKeyb12		; CLK IN = LOW -> loop
	bsf		KStatus, KBTX	; segnala trasmissione completata
	bcf		T0CON, TMR0ON	; stop Timer 0
	bcf		INTCON, TMR0IF	; azzera flag timeout
	return

GetKbd:
	call	RxKeyb			; ricezione entro timeout
	btfsc	KStatus, KRESET	; test reset
	reset					; ricevuto reset in -> reset
	btfsc	KStatus, KTIMEO	; test timeout
	bra		GetKbd10		; errore ricezione
	ldw		KStatus			; test errore ricezione
	andlw	KBRXERR
	bnz		GetKbd10		; dato ricevuto errato
	bcf		STATUS, C		; C = = -> OK
	return
GetKbd10:
	bsf		STATUS, C		; C = 1 -> errore
	return

SendLed:
	movlw	KBD_SETLED
	rcall	SendKbd
	bc		SendKbd10
	ldw		LedSt
	
SendKbd:
	rcall	TxKeyb
	btfsc	KStatus, KRESET	; test reset
	reset					; ricevuto reset in -> reset
	btfsc	KStatus, KTIMEO	; test timeout
	bra		SendKbd10		; errore
	btfss	KStatus, KBTX	; test trasmissione OK
	bra		SendKbd10		; trasmissione fallita - errore
	btfss	KStatus, KBTX	; test trasmissione OK
	bra		SendKbd10		; trasmissione fallita - errore
	btfsc	KStatus, KBACK	; test errore ACK
	bra		SendKbd10		; errore ACK
	rcall	GetKbd			; ricezione da tastiera
	bc		SendKbd10		; errore ricezione
	movlw	KBD_ACK			; test ACK tastiera
	cpfseq	KBuf
	bra		SendKbd10		; errore
	bcf		STATUS, C		; C = = -> OK
	return
SendKbd10:
	bsf		STATUS, C		; C = 1 -> errore
	return

InitIO:
	bcf		RCON, 6
	bsf		RCON, 1
	clrf	INTCON			; disabilita interrupt
	clrf	INTCON2			; RB0, RB1, RB2 int su fronte negativo
	clrf	INTCON3
	clrf	PIR1
	clrf	PIR2
	clrf	PIE1
	clrf	PIE2
	clrf	ADCON0			; disabilita modulo A/D
	setf	DDRA			; tutte le porte in input
	setf	DDRB
	setf	DDRC
	setf	DDRD
	setf	DDRE
	movlw	B'00001111'		; tutte le porte I/O sono digitali
	movwf	ADCON1
	setf	LATA			; tutte le porte I/O a livello alto
	setf	LATB
	setf	LATC
	setf	LATD
	clrf	LATE			; porta RE a livello basso in uscita
	movlw	B'11000011'		; RA<0:1> input, le altre output
	movwf	DDRA
	movlw	B'11001111'		; RB<4:5> in output, le altre in input
	movwf	DDRB
	clrf	DDRC			; porta C in uscita
	bcf		LATC, KBVOFF	; resetta alimentazione tastiera
	clrf	DDRE			; porta E in uscita
	clrf	TBLPTRU			; FLASH solo < 64K
	rcall	InitData		; inizializza variabili

Init1:						; attesa linea reset in alta
	btfss	PORTB, RESIN
	bra		Init1

	bsf		RCON, IPEN		; abilita int. priorizzati
	bcf		INTCON2, RBPU	; disabilita pull-up su porta B
	bsf		INTCON3, INT2IP	; alta priority per INT2 su RB2 (reset in)
	bcf		INTCON3, INT1IP	; bassa priority per INT1 su RB1 (linea DRQ)
	bsf		INTCON, INT0IE	; abilita INT0 su RB0 high priority (DATA IN)
	bsf		INTCON3, INT1IE	; abilita INT1 su RB1 (linea DRQ)
	bsf		INTCON3, INT2IE	; abilita INT2 su RB2 (linea reset in)
	bsf		LATC, KBVOFF	; termina impulso negativo VCC tastiera
	movlw	LOW(T100US)		; imposta ritardo 100us
	movwf	T0CntL
	movlw	HIGH(T100US)
	movwf	T0CntH
	rcall	Delay_us		; ritardo 100us
	bcf		LATC, KBVON		; attiva alimentazione tastiera
	movlw	LOW(T20US)		; imposta ritardo 20us
	movwf	T0CntL
	movlw	HIGH(T20US)
	movwf	T0CntH
	rcall	Delay_us		; ritardo 20us
	bsf		LATC, KBVON		; termina impulso negativo
	movlw	LOW(T40MS)		; imposta ritardo 40ms per stabilizzare VCC 
	movwf	T0CntL
	movlw	HIGH(T40MS)
	movwf	T0CntH
	rcall	Delay_ms		; ritardo 40ms
	bcf		INTCON, INT0IF	; resetta flag INT0 (eventuali impulsi spuri)
	movlw	TMR0P256		; TMR0 stop, prescaler = 256, modo 16 bit
	movwf	T0CON	
	bcf		INTCON, TMR0IF	; azzera flag timeout
	movlw	HIGH(T1500MS)	; imposta timeout 1500ms
	movwf	TMR0H			; caricare prima TMR0H
	movlw	LOW(T1500MS)
	movwf	TMR0L			; carica TMR0L/H
	bsf		T0CON, TMR0ON	; start Timer 0

Init2:						; loop attesa tastiera entro timeout 1500ms
	btfsc	INTCON, TMR0IF
	bra		Init4			; timeout
	btfsc	INTCON3, INT2IF
	reset					; INT2 -> reset
	btfss	INTCON, INT0IF
	bra		Init2			; loop
	bcf		T0CON, TMR0ON	; stop Timer 0
	bcf		INTCON, TMR0IF	; azzera flag timeout
	call	RxKeyb			; INT0 -> riceve dato da tastiera
	btfsc	KStatus, KRESET	; test reset
	reset					; ricevuto reset in -> reset
	btfsc	KStatus, KTIMEO	; test timeout
	bra		Init4			; scarta dato ricevuto se timeout
	btfss	KStatus, KBRX	; test ricezione full
	bra		Init4			; ricezione incompleta -> scarta dato
	ldw		KStatus			; test errore ricezione
	andlw	KBRXERR
	bnz		Init4			; dato ricevuto errato
	movlw	KBD_BAT			; test BAT OK
	cpfseq	KBuf
	bra		KNoBAT			; no BAT OK !
	bsf		KbdFlag, KBBAT	; imposta flag BAT OK
	bsf		KbdFlag, KBBAT2	; imposta flag BAT OK
	bra		Init3

KNoBAT:
	movlw	KBD_NOBAT		; test BAT failure
	cpfseq	KBuf
	bra		Init3
	bsf		KbdFlag2, KBNOBAT	; imposta flag BAT FAILURE

Init3:
	bsf		KbdFlag, KBFND	; imposta flag tastiera presente

Init4:		; KBFND indica tastiera prsente - se il flag e' resettato
			; occorre fare ulteriore verifica della tastiera
	bcf		T0CON, TMR0ON	; stop Timer 0
	bcf		INTCON, TMR0IF	; azzera flag timeout
	bcf		INTCON, INT0IF	; resetta flag INT0
	btfss	KbdFlag, KBFND	; se tastiera presente salta ulteriore verifica
	rcall	KbVerf			; verifica tastiera
	bcf		INTCON, INT0IF	; resetta flag INT0
	btfss	KbdFlag, KBFND	; se tastiera presente la inizializza
	bra		Init10			; altrimenti esce
	btfss	KbdFlag, KBBAT	; tastiera ok ?
	bra		Init10			; no -> esce

	; a questo punto tastiera presente e funzionante
	; si legge ID tastiera
	bcf		INTCON, INT0IF	; resetta flag INT0
	movlw	KBD_GETID
	call	SendKbd
	bc		Init6			; errore
	call	GetKbd			; riceve ID LOW
	bc		Init6			; errore
	ldw		KBuf
	movwf	KbdIDL			; salva ID LOW
	call	GetKbd			; riceve ID LOW
	bc		Init6			; errore
	ldw		KBuf
	movwf	KbdIDH			; salva ID HIGH
	movlw	KBD_IDL			; test ID
	cpfseq	KbdIDL
	bra		Init6			; ID LOW errato
	movlw	KBD_IDH			; test ID
	cpfseq	KbdIDH
	bra		Init6			; ID HIGH errato
	bsf		KbdFlag, KBVID	; imposta flag ID valido

Init6:
	bcf		INTCON, INT0IF	; resetta flag INT0
	movlw	KBD_SETLED		; spegne LED
	call	SendKbd
	bc		Init61
	movlw	0x00
	call	SendKbd
	bc		Init61
	bsf		KbdFlag, KBLEDOK

Init61:
	bcf		INTCON, INT0IF	; resetta flag INT0
	movlw	KBD_SETSC		; imposta scancode 02
	call	SendKbd
	bc		Init62
	movlw	0x02
	call	SendKbd
	bc		Init62
	bsf		KbdFlag, KBSCOK

Init62:
	bcf		INTCON, INT0IF	; resetta flag INT0
	movlw	KBD_TYPERATE	; imposta typerate
	call	SendKbd
	bc		Init63
	movlw	KBD_DEFRATE
	call	SendKbd
	bc		Init63
	bsf		KbdFlag, KBRATE

Init63:
	bcf		INTCON, INT0IF	; resetta flag INT0
	movlw	KBD_ENABLE		; abilita scanning tastiera
	call	SendKbd
	bc		Init10
	bsf		KbdFlag, KBENABLE

Init10:
	SCLOCKL					; disabilita TX da tastiera
	movlw	0x03			; registro KFR 3
	movwf	LATE
	ldw		KbdFlag
	call	KFRWrite
	bcf		INTCON, INT0IF	; resetta flag INT0
	bcf		INTCON3, INT1IF	; resetta flag INT1
	bcf		INTCON3, INT2IF	; resetta flag INT2
	bsf		INTCON, GIEH	; abilita INT high priority
	bsf		INTCON, GIEL	; abilita INT low priority
	return

InitData:
	lfsr	FSR0, 0x0		; azzera RAM banco 0, 1 & 2
	movlw	0x03			; banco 3 non esiste !

InitData1:
	clrf	POSTINC0
	cpfseq	FSR0H
	bra		InitData1

	movlw	0x04			; contatore registri
	movwf	KFRRdCnt
	clrf	LATE			; azzera indirizzo KFR
	clrf	DDRD			; porta D in uscita
	clrf	LATD			; porta D tutti i bit bassi

InitData2:
	nop						; scrittura su registro
	nop
	bcf		LATC, KFRWE		; impulso write
	nop
	bsf		LATC, KFRWE
	nop
	incf	LATE,f			; incrementa indirizzo
	decfsz	KFRRdCnt, f		; loop
	bra		InitData2
	lfsr	FSR2, RxBuffer	; imposta puntatore iniziale testa buffer RX
	lfsr	FSR1, RxBuffer	; imposta puntatore iniziale coda buffer RX
	movlw	LOW(TxBuffer)		; imposta puntatore Head, Tail TX a inizio buffer
	movwf	TxHeadL
	movwf	TxTailL
	movlw	HIGH(TxBuffer)
	movwf	TxHeadH
	movwf	TxTailH
	return

Delay_ms:
	movlw	TMR0P4			; TMR0 stop, prescaler = 4, modo 16 bit
	bra		Delay_xx

Delay_us:
	movlw	TMR0P0			; TMR0 stop, no prescaler, modo 16 bit
Delay_xx:
	movwf	T0CON	
	bcf		INTCON, TMR0IF	; azzera flag timeout
	ldw		T0CntH			; imposta timeout high
	movwf	TMR0H			; caricare prima TMR0H
	ldw		T0CntL
	movwf	TMR0L			; carica TMR0L/H
	bsf		T0CON, TMR0ON	; start Timer 0

Delay_xx1:					; loop
	btfss	INTCON, TMR0IF	; attende flag overflow
	bra		Delay_xx1		; loop ritardo
	bcf		T0CON, TMR0ON	; stop Timer 0
	bcf		INTCON, TMR0IF	; azzera flag timeout
	return

KbVerf:
	movlw	KBD_ECHO		; test tastiera tramite ECHO
	call	TxKeyb
	btfsc	KStatus, KRESET	; test reset
	reset					; ricevuto reset in -> reset
	btfsc	KStatus, KTIMEO	; test timeout
	return					; tastiera non presente
	btfss	KStatus, KBTX	; test trasmissione OK
	return					; trasmissione fallita - si assume NO tastiera
	call	RxKeyb			; ricezione ECHO entro timeout
	btfsc	KStatus, KRESET	; test reset
	reset					; ricevuto reset in -> reset
	btfsc	KStatus, KTIMEO	; test timeout
	return					; tastiera non presente
	bsf		KbdFlag, KBFND	; si assume tastiera presente
	btfss	KStatus, KBRX	; test ricezione full
	return					; ricezione incompleta -> failure tastiera
	ldw		KStatus			; test errore ricezione
	andlw	KBRXERR
	bnz		KbVerf2			; dato ricevuto errato -> failure tastiera
	movlw	KBD_ECHO		; test ECHO
	cpfseq	KBuf
	return					; failure tastiera
	bsf		KbdFlag, KBBAT	; imposta flag BAT OK
KbVerf2:
	return

KbdProc:
	movlw	0x00				; test presenza bytes nella coda di ricezione
	subwf	RxCnt, W
	bnz		KbdProc0			; coda non vuota
	bcf		RxFlag, RXOVR		; rimuove blocco TX (coda vuota)
	SCLOCKH						; abilita TX da tastiera
	return
KbdProc0:
	btfsc	TxFlag, TXOVR		; buffer TX pronto ?
	return						; NO - attende posto
	ldw		POSTINC1			; legge byte da coda -> FSR1 ptr coda di ricezione
	movwf	RxData
	bcf		INTCON, GIEH		; disabilita INT
	bcf		INTCON, GIEL
	decf	RxCnt, F			; aggiorna contatore bytes in coda
	bcf		RxFlag, RXOVR		; rimuove blocco TX (coda vuota)
	SCLOCKH						; abilita TX da tastiera	
	bsf		INTCON, GIEH		; abilita INT
	bsf		INTCON, GIEL
	; si assume che TBLPTR non cambia entro il ciclo
	btfss	RxFlag, RXEXT1		; sequenza E1 pending ?
	bra		KbdProc1			; NO
	tblrd	*+					; legge byte da memoria ed incrementa ptr
	ldw		TABLAT
	subwf	RxData, W			; confronta con dato ricevuto
	btfss	STATUS, Z			; Z = 1 se coincide
	bsf		RxFlag, RXE1ERR		; segnala errore in sequenza E1
	decfsz	E1Cnt, F			; decrementa contatore
	bra		UpdBufPtr
E1Test:
	btfss	RxFlag, RXE1ERR		; errore nella sequenza E1 ?
	bra		PauseKey			; NO - tasto PAUSE premuto
ClrRX:							; termina scansione corrente - inizio nuova
	movlw	RXMASK				; azzera tutti i flag ricezione
	andwf	RxFlag, F
UpdBufPtr:
	movlw	LOW(RxBufEnd)		; coda circolare - test puntatore
	subwf	FSR1L, W			; FSR1L - LOW(RxBufEnd)
	movlw	HIGH(RxBufEnd)		; FSR1H - HIGH(RxBufEnd)
	subwfb	FSR1H, W			; C = 0 se FSR1 < RxBufEnd
	btfsc	STATUS, C			; OK
	lfsr	FSR1, RxBuffer		; reimposta puntatore a inizio buffer
	return
PauseKey:
	movlw	PAUSECODE			; imposta codice tasto PAUSE
	movwf	RxCode
	rcall	GetKeyTbl
	bra		KbdProc14			; gestione tasti toggle
KbdProc1:
	movlw	KBD_BRK				; test break
	subwf	RxData, W
	bnz		KbdProc2
	bsf		RxFlag, RXBRK		; F0 break pending
	bra		UpdBufPtr
KbdProc2:
	movlw	KBD_EXT				; test codice esteso
	subwf	RxData, W
	bnz		KbdProc3
	movlw	RXMASK				; azzera tutti i flag ricezione
	andwf	RxFlag, F
	bsf		RxFlag, RXEXT		; imposta E0 pending
	bra		UpdBufPtr
KbdProc3:
	movlw	KBD_EXT1			; test codice esteso
	subwf	RxData, W
	bnz		KbdProc4
	movlw	RXMASK				; azzera tutti i flag ricezione
	andwf	RxFlag, F
	bsf		RxFlag, RXEXT1		; imposta E1 pending
	movlw	LOW EXT1TBL			; imposta ptr tabella sequenza E1
	movwf	TBLPTRL
	movlw	HIGH EXT1TBL
	movwf	TBLPTRH
	movlw	E1SEQCNT			; imposta numero bytes sequenza E1
	movwf	E1Cnt
	bra		UpdBufPtr
KbdProc4:
	btfss	RxFlag, RXEXT		; codice esteso E0 pending ?
	bra		KbdProc6			; NO - codice normale o break
	movlw	LOW EXTTBL			; imposta ptr tabella sequenza E0
	movwf	TBLPTRL
	movlw	HIGH EXTTBL
	movwf	TBLPTRH
	setf	E1Cnt				; indice tabella E0 (inizializzato a 0xFF)
KbdProc5:
	incf	E1Cnt, F			; update indice
	tblrd	*+					; legge byte da memoria ed incrementa ptr
	movlw	0xFF				; test fine tabella
	subwf	TABLAT, W			
	bz		ClrRX				; termina sequenza
	ldw		TABLAT
	subwf	RxData, W			; confronta con dato ricevuto
	bnz		KbdProc5			; loop next
								; trovato codice esteso
	movlw	LOW EXTRLTBL		; imposta ptr tabella rilocazione E0
	addwf	E1Cnt, W			; aggiunge indice
	movwf	TBLPTRL
	movlw	HIGH EXTRLTBL
	movwf	TBLPTRH
	clrf	WREG
	addwfc	TBLPTRH, F
	tblrd	*+					; codice rilocato
	ldw		TABLAT
	bra		KbdProc10
KbdProc6:						; codice normale
	movlw	LOW DEFTBL			; imposta ptr tabella sequenza normale
	movwf	TBLPTRL
	movlw	HIGH DEFTBL
	movwf	TBLPTRH
	setf	E1Cnt				; indice tabella (inizializzato a 0xFF)
KbdProc7:
	incf	E1Cnt, F			; update indice
	tblrd	*+					; legge byte da memoria ed incrementa ptr
	movlw	0xFF				; test fine tabella
	subwf	TABLAT, W			
	bz		KbdProc8			; codice da non rilocare
	ldw		TABLAT
	subwf	RxData, W			; confronta con dato ricevuto
	bnz		KbdProc7			; loop next
								; trovato codice da rilocare
	movlw	LOW DEFRLTBL		; imposta ptr tabella rilocazione 
	addwf	E1Cnt, W			; aggiunge indice
	movwf	TBLPTRL
	movlw	HIGH DEFRLTBL
	movwf	TBLPTRH
	clrf	WREG
	addwfc	TBLPTRH, F
	tblrd	*+					; codice rilocato
	ldw		TABLAT
	bra		KbdProc10
KbdProc8:						; rilocazione degli scan code 70..7E in 60..6E
	movlw	0x70
	subwf	RxData, W
	ldw		RxData				; W = codice non rilocato
	bnc		KbdProc10			; OK (< 70)
	movlw	0x7F				; max. codice di scansione 0x7E
	subwf	RxData, W
	bc		ClrRX				; codice non valido (>= 7F) - azzera ed esce
	movlw	0x10				; rilocazione codice in 60..6E
	subwf	RxData, W			; W = codice rilocato
KbdProc10:						; IN -> W = codice
	movwf	E1Cnt				; codice
	movlw	LOW CVTTBL			; imposta ptr tabella conversione 
	addwf	E1Cnt, W			; aggiunge indice
	movwf	TBLPTRL
	movlw	HIGH CVTTBL
	movwf	TBLPTRH
	clrf	WREG
	addwfc	TBLPTRH, F
	tblrd	*+					; codice rilocato
	ldw		TABLAT
	movwf	RxCode				; nuovo codice
	bmi		ClrRX				; codice non valido - azzera flag ed esce
	movlw	('z' + 1)			; massimo codice 7A ('z')
	subwf	RxCode, W
	bc		ClrRX				; codice non valido

	; RxCode contiene ora un codice compreso tra 00 e 7A
	; codici 00 - 41 utilizzati per accesso tabella KEYTBL
	; codici 42 - 56 utilizzati per accesso mappe US/IT
	; codici 57 - 60 non validi
	; codici 61 - 7A lettere 'a'..'z'
	; codici 00 - 05 -> flag shift/alt/ctrl
	; codici 06 - 09 -> tasti toggle (LOCK e PAUSE)	

	bcf		TxFlag, SKPCTLB		; azzera flag skip SHIFT/ALT/CTL
	bcf		TxFlag, SKPALTB
	bcf		TxFlag, SKPSHFB		
	bcf		MapCfg, MAPALPHA	; tasto non alfabetico	
	movlw	'a'					; test carattere alfabetico
	subwf	RxCode, W
	bnc		MKeyTst				; no alpha - test codice mappe US/IT
	clrf	TxData0				; flag codice ASCII
	ldw		RxCode				; 'a'..'z'
	movwf	TxData1				; codice ASCII tasto
	movlw	0x20				; converte in maiuscolo
	subwf	RxCode, W			; 'A'..'Z'
	movwf	Caps
	bsf		MapCfg, MAPALPHA	; tasto alfabetico	
	bra		KbdProc40			; gestione tasto alpha

MKeyTst:
	movlw	MAPCODE				; test codice in 42..56 (mappe US/IT)
	subwf	RxCode, W
	bnc		NKeyTst				; CODE < MAPCODE - no tasto mappe US/IT
	movwf	RxData				; salva CODE - MAPCODE (indice 0)
	movlw	(MAPCODEMAX + 1)
	subwf	RxCode, W
	bnc		MapTst				; CODE < MAPCODEMAX + 1 - tasto mappe US/IT
	bra		ClrRX				; codice non valido
MapTst:
	; gestione mappe US/IT
	bcf		STATUS, C			; azzera carry
	rlcf	RxData, f			; * 2
	bsf		TxFlag, SKPSHFB		; skip test SHIFT
	btfsc	MapCfg, MAPITB		; configurazione mappa IT ?
	bra		ITMap				; SI
	btfsc	ShfSt, ASHFB		; SHIFT ?
	incf	RxData, F			; SI - incrementa indice
	movlw	LOW USTBL			; imposta ptr tabella conversione mappa US 
	addwf	RxData, W			; aggiunge indice
	movwf	TBLPTRL
	movlw	HIGH USTBL
	movwf	TBLPTRH
	clrf	WREG
	addwfc	TBLPTRH, F
	tblrd	*					; codice tasto
	ldw		TABLAT
	movwf	TxData1				; codice tasto
	clrf	TxData0				; flag ASCII
	bra		KbdProc41			; gestione tasto
ITMap:
	movlw	0x00
	movwf	TxData1
	clrf	TxData0
	bra		KbdProc41			; gestione tasto

NKeyTst:
	; gestione tasti NO-REPEAT: PWR, SLEEP, WAKEUP, APP
	; L-GUI, R-GUI, ESC, PRINT SCRN
	movlw	WAKECODE
	subwf	RxCode, W
	bnc		NMITst				; RxCode < WAKECODE
	movwf	RxData				; RxData = RxCode - WAKECODE
	movlw	(ESCCODE + 1)
	subwf	RxCode, W
	bc		NMITst				; RxCode >= (ESCCODE + 1)
	movlw	LOW RPTTBL			; imposta ptr tabella flag RPTTBL 
	addwf	RxData, W			; aggiunge indice
	movwf	TBLPTRL
	movlw	HIGH RPTTBL
	movwf	TBLPTRH
	clrf	WREG
	addwfc	TBLPTRH, F
	tblrd	*					; flag tasto
	ldw		TABLAT
	movwf	RxData				; flag	
	btfsc	RxFlag, RXBRK		; test tasto rilasciato
	bra		NKeyRls				; SI
	andwf	RptSt, W			; flag tasto premuto attivo ?
	bz		NKeySet				; NO - prima attivazione
	bra		ClrRX				; tasto repeat - scarta codice
NKeySet:
	ldw		RxData				; set bit in RptSt
	iorwf	RptSt, F
	bnz		NMITst
NKeyRls:
	comf	RxData, W			; complementa maschera per azzerare bit
	andwf	RptSt, F			; azzera bit

NMITst:
	; CTRL + BREAK, ALT + SYSREQ, WAKE UP, PWR, SLEEP sono tasti NMI
	; vengono rilevati solo dopo il loro rilascio
	rcall	GetKeyTbl
	movf	TxData1, F			; test se codice = 0
	bnz		NMITst1	
	bra		ClrRX				; tasto nullo - scarta codice
NMITst1:	
	movlw	0xC0				; test tasto NMI
	subwf	TxData0, W			; flag tasto
	bnz		KbdProc12			; no tasto NMI
	btfss	RxFlag, RXBRK		; codice break ?
	bra		ClrRX				; NO - scarta codice
	bcf		RxFlag, RXBRK		; azzera flag
	bra		KbdProc60			; gestione tasto NMI

KbdProc12:
	; gestione tasti SHIFT/ALT/CTRL
	; TxData0 = flag tasto
	; TxData1 = codice tasto
	movlw	0x80				; test invio status flag
	subwf	TxData0, W
	bnz		KbdProc14			; no SHIFT/ALT/CTRL
	ldw		CtlSt				; salva flag attuali
	movwf	TCtlSt
	movlw	(ASHFM | ALTCTLM)	; maschera bit necessari
	andwf	ShfSt, W
	movwf	TShfSt
	ldw		TxData1				; flag tasto
	movwf	RxData				; flag attuale
	comf	RxData, W			; complementa maschera per azzerare bit
	andwf	CtlSt, F			; azzera bit
	ldw		RxData				; carica maschera bit attivi
	btfss	RxFlag, RXBRK		; lascia bit azzerati se tasto rilasciato
	iorwf	CtlSt, F			; tasto premuto: imposta bit
	bcf		CtlSt, AALTB		; azzera bit di ALT e CTRL
	bcf		CtlSt, ACTLB
	movlw	ALTKM				; test L-ALT OR R-ALT
	andwf	CtlSt, W
	btfss	STATUS, Z			; Z = 1 -> no ALT 
	bsf		CtlSt, AALTB		; imposta bit ALT (L o R)
	movlw	CTRLKM				; test L-CTRL OR R-CTRL
	andwf	CtlSt, W
	btfss	STATUS, Z			; Z = 1 -> no CTRL
	bsf		CtlSt, ACTLB		; imposta bit CTRL (L o R)
	bcf		ShfSt, ASHFB		; azzera bit SHIFT
	bcf		ShfSt, ALTCTLB		; azzera bit ALT OR CTRL
	movlw	SHIFTKM				; test L-SHIFT OR R-SHIFT
	andwf	CtlSt, W
	btfss	STATUS, Z			; Z = 1 -> no SHIFT
	bsf		ShfSt, ASHFB		; imposta bit SHIFT (L o R)
	movlw	ALLALTCTLM			; test ALT OR CTRL
	andwf	CtlSt, W
	btfss	STATUS, Z			; Z = 1 -> no ALT OR CTRL
	bsf		ShfSt, ALTCTLB		; imposta bit ALT OR CTRL (L o R)
	ldw		CtlSt				; confronta flag attuali con flag precedenti
	cpfseq	TCtlSt
	bra		KbdProc13			; flag cambiato
	movlw	(ASHFM | ALTCTLM)
	andwf	ShfSt, W			; maschera bit necessari
	subwf	TShfSt, W			; confronta con vecchio flag
	bnz		KbdProc13			; flag cambiato
	bra		ClrRX				; flag non cambiati -> scarta codice
KbdProc13:						; invia nuovi flag all' host
	; formato -> 100000<76-ShfSt> + CtlSt
	ldw		CtlSt
	movwf	TxData1
	ldw		ShfSt				; trasla bit 7 e 6 in bit 1 e 0
	movwf	RxData
	rrncf	RxData, F			; bit 6
	rrncf	RxData, F			; bit 5
	rrncf	RxData, F			; bit 4
	rrncf	RxData, F			; bit 3
	rrncf	RxData, F			; bit 2
	rrncf	RxData, W			; bit 1 -> W
	andlw	B'00000011'
	iorwf	TxData0, F

PutKey:							; inserisce word codice tasto nella coda TX
	movlw	(TXMAXCNT - 1)		; test spazio disponibile
	subwf	TxCnt, W			; W = TxCnt - (TMAXCNT - 1)
	btfsc	STATUS, C			; W >= 0 -> C = 1
	bsf		TxFlag, TXOVR		; segnala overrun buffer TX
	ldw		TxHeadL				; imposta ptr head al buffer
	movwf	FSR0L
	ldw		TxHeadH
	movwf	FSR0H
	ldw		TxData0				; primo byte in coda
	movwf	POSTINC0
	ldw		TxData1				; secondo byte in coda
	movwf	POSTINC0
	incf	TxCnt, F			; aggiorna contatore word nel buffer TX
	ldw		FSR0L				; aggiorna ptr
	movwf	TxHeadL
	ldw		FSR0H
	movwf	TxHeadH	
	movlw	LOW(TxBufEnd)		; coda circolare - test puntatore
	subwf	TxHeadL, W			; TxHeadL - LOW(TxBufEnd)
	movlw	HIGH(TxBufEnd)		; TxHeadH - HIGH(TxBufEnd)
	subwfb	TxHeadH, W			; C = 0 se TxHead < TxBufEnd
	btfss	STATUS, C			; test buffer
	bra		ClrRX				; C = 0 -> OK
	movlw	LOW(TxBuffer)		; reimposta puntatore a inizio buffer
	movwf	TxHeadL
	movlw	HIGH(TxBuffer)
	movwf	TxHeadH
	bra		ClrRX				; azzera flag ed esce
 
KbdProc14:
	; gestione tasti TOGGLE
	movlw	0x90				; test invio toggle
	subwf	TxData0, W			; test tasti LOCK o PAUSE
	bnz		KbdProc16
	ldw		TglSt				; salva flag toggle attuali
	movwf	TTglSt
	ldw		LockSt
	movwf	TLockSt
	movlw	PAUSECODE			; test PAUSE
	subwf	RxCode, W
	bnz		NoPause				; no tasto PAUSE
	btg		TglSt, PAUSB		; toggle bit ad ogni pressione PAUSE
	bra		KbdProc15
NoPause:
	btfsc	RxFlag, RXBRK		; test tasto rilasciato
	bra		ToggleRls			; SI
	andwf	LockSt, W			; flag tasto premuto attivo ?
	bz		ToggleBit			; NO - prima attivazione
	bra		ClrRX				; tasto repeat - scarta codice
ToggleBit:
	ldw		TxData1				; toggle bit in TglSt
	xorwf	TglSt, F
	iorwf	LockSt, F			; setta bit in LockSt
	bnz		TstLock
ToggleRls:
	comf	TxData1, W			; complementa maschera per azzerare bit
	andwf	LockSt, F			; azzera bit
TstLock:
	ldw		TglSt				; confronta con toggle precedente
	subwf	TTglSt, W
	bnz		LedUpd
	bra		ClrRX				; toggle non modificato - scarta codice
LedUpd:
	call	SetLed1
	btfss	KbdFlag, KBLEDOK	; comando LED riuscito
	bra		ClrRX				; ignora tasto ed esce
KbdProc15:						; invia toggle ad host
	; formato -> 10010000 + TglSt
	ldw		TglSt
	movwf	TxData1
	bra		PutKey
	
KbdProc16:
	; gestione tasti tastierino numerico (KP)
	movlw	KP0CODE
	subwf	RxCode, W
	bnc		KbdProc20			; no tasto numerico
	movlw	INSCODE
	subwf	RxCode, W
	bc		KbdProc20			; no tasto numerico
	; se NUM-LOCK OFF i tasti numerici diventano tasti cursore (escluso 5)
	btfss	TglSt, NUMLB
	bra		KbdProc18			; NUM-LOCK OFF -> converte in codice cursore
	; se NUM-LOCK ON e premuto il solo tasto SHIFT converte in codice cursore
	btfss	ShfSt, ASHFB
	bra		KbdProc20			; no SHIFT -> tasto numerico
	btfsc	ShfSt, ALTCTLB		; premuti ALT e/o CTRL ?
	bra		KbdProc20			; si - tasto numerico
	bsf		TxFlag, SKPSHFB		; flag skip test SHIFT
	bsf		TxFlag, SKPALTB		; flag skip test ALT
	bsf		TxFlag, SKPCTLB		; flag skip test CTRL

KbdProc18:
	movlw	CRSLEN				; converte in codice tasto cursore
	addwf	RxCode, F
	rcall	GetKeyTbl			; legge nuovamente codice/flag tasto
	movf	TxData1, F			; test se codice tasto = 0
	bnz		KbdProc20
	bra		ClrRX				; scarta codice nullo

KbdProc20:
	; toggle tasto INSERT - soltanto senza SHIFT/ALT/CTRL
	movlw	INSCODE
	subwf	RxCode, W
	bnz		KbdProc22
	btfsc	TxFlag, SKPSHFB		; INSERT su tastierino numerico ?
	bra		KbdProc21			; si - non testa SHIFT	
	movlw	SHFALL
	andwf	ShfSt, W			; test SHIFT/ALT/CTRL
	bnz		KbdProc22			; no toggle
KbdProc21:
	btfsc	RxFlag, RXBRK		; test tasto rilasciato
	bra		KbdProc22
	btg		TglSt, INSTB		; toggle bit INSERT

KbdProc22:
	; imposta flag in TglSt per ESC e PRINT SCRN
	movlw	PRTCODE
	subwf	RxCode, W
	bz		KbdProc23
	movlw	ESCCODE
	subwf	RxCode, W
	bnz		KbdProc24
KbdProc23:
	movlw	ESCPRNM				; imposta flag ESC e PRINT SCRN 
	andwf	RptSt, W			; da RptSt a TglSt
	iorwf	TglSt, F
	
KbdProc24:
	; gestione tasti CR, TAB, SHIFT-CR, SHIFT-TAB, '<' e '>' mappa IT
	movlw	CRCODE
	subwf	RxCode, W
	bnc		KbdProc26			; RxCode < CRCODE
	movlw	KPBARCODE
	subwf	RxCode, W
	bc		KbdProc26			; RxCode >= KPBARCODE
	btfss	ShfSt, ASHFB		; SHIFT ?
	bra		KbdProc25			; NO
	btfsc	ShfSt, ALTCTLB		; CTRL/ALT ?
	bra		KbdProc25			; SI
	incf	RxCode, F			; SHIFT -> increnmenta codice
KbdProc25:
	rcall	GetKeyTbl			; legge nuovamente codice/flag tasto
	movf	TxData1, F			; test se codice tasto = 0
	bnz		KbdProc50			; gestione tasto
	bra		ClrRX				; scarta codice nullo

KbdProc26:
	; tasti funzione - a questo punto CODE < MAPCODE (42)
	movlw	FNCCODE				; range tasti funzione 36..41
	subwf	RxCode, W
	bnc		KbdProc50			; gestione tasti per codici < FNCCODE (36)

	;						FX -> A0..AB
	; SHIFT +				FX -> AC..B7
	; CTL	+				FX -> B8..C3
	; ALT	+				FX -> C4..CF
	; CTL + SHIFT +			FX -> D0..DB
	; CTL + ALT	 +			FX -> DC..E7
	; ALT + SHIFT +			FX -> E8..F3
	; CTL + ALT + SHIFT	+	FX -> F4..FF
	
	btfss	ShfSt, ASHFB
	bra		FnNoShf
	movlw	D'12'				; SHIFT + Fx
	btfss	ShfSt, ALTCTLB		; ALT o CTRL ?		
	bra		FnAdd				; no - SHIFT + FX
	movlw	D'48'
	btfss	CtlSt, AALTB		; ALT ?
	bra		FnAdd				; no - SHIFT + CTRL + Fx
	movlw	D'72'	
	btfss	CtlSt, ACTLB		; CTRL ?
	bra		FnAdd				; no - SHIFT + ALT + Fx
	movlw	D'84'				; SHIFT + ALT + CTRL + Fx
	bra		FnAdd
FnNoShf:
	movlw	D'0'				; Fx
	btfss	ShfSt, ALTCTLB		; ALT o CTRL ?		
	bra		FnAdd				; no - FX
	movlw	D'24'
	btfss	CtlSt, AALTB		; ALT ?
	bra		FnAdd				; no - CTRL + Fx
	movlw	D'36'	
	btfss	CtlSt, ACTLB		; CTRL ?
	bra		FnAdd				; no - ALT + Fx
	movlw	D'60'				; ALT + CTRL + Fx
FnAdd:
	addwf	TxData1, F
	bra		KeySnd				; tasti funzione inviati senza altri test

KbdProc40:						; gestione tasto alfabetico
	ldw		Caps				; 'A'..'Z'
KbdProc41:						; gestione tasto ASCII non-alpha
	movwf	RxData				; codice tasto attuale (TxData1)
	movlw	'@'
	subwf	RxData, W
	bnc		KbdProc50			; key < '@' (40)
	movlw	'`'
	subwf	RxData, W
	bc		KbdProc50			; key >= '`' (60)
	btfss	CtlSt, AALTB		; ALT ?
	bra		KbdProc42			; NO
	movlw	0x1F
	andwf	RxData, W			; converte in 00..1F
	addlw	0x60				; converte in 60..7F
	movwf	TxData1				; codice controllo
	movlw	0xA0				; flag codice controllo
	movwf	TxData0
	bra		KbdProc50
KbdProc42:
	btfss	CtlSt, ACTLB		; CTRL ?
	bra		KbdProc43			; NO
	movlw	0x1F
	andwf	RxData, W			; converte in 00..1F
	movwf	TxData1				; codice ASCII
	clrf	TxData0				; flag codice ASCII
	bra		KbdProc50
KbdProc43:
	; test SHIFT / CAPS-LOCK per tasti alfabetici
	btfss	MapCfg, MAPALPHA
	bra		KbdProc50			; no tasto alfabetico
	ldw		ShfSt				; SHIFT
	xorwf	TglSt, W			; confronta con CAPS LOCK
	andlw	0x80				; interessa bit 7 soltanto
	bz		KbdProc50			; o NO SHIFT oppure CAPS LOCK + SHIFT -> a..z
	ldw		Caps				; o SHIFT oppure CAPS LOCK -> A..Z
	movwf	TxData1
			
KbdProc50:
	; ALT 0..9        -> 30..39
	; CTRL 0..9       -> 40..49
	; CTRL + ALT 0..9 -> 50..59
	btfsc	TxData0, 7
	bra		Key10				; codice di controllo
	movlw	'0'					; test tasto numerico
	subwf	TxData1, W
	movwf	RxData				; 0..9
	bnc		Key06
	movlw	('9' + 1)
	subwf	TxData1, W
	bc		Key06
	btfss	CtlSt, AALTB		; ALT ?
	bra		Key01				; NO
	movlw	ALTKEY0				; ALT + 0..9
	btfsc	CtlSt, ACTLB		; CTRL + ALT ?
	movlw	ALTCTLKEY0			; CTRL + ALT + 0..9
	bra		Key02
Key01:
	btfss	CtlSt, ACTLB		; CTRL ?
	bra		KeySnd				; NO - lascia inalterato il tasto
	movlw	CTLKEY0				; CTRL + 0..9
Key02:
	addwf	RxData, W
	movwf	TxData1				; codice controllo
	movlw	0xA0				; flag codice controllo
	movwf	TxData0
	bra		KeySnd				; invia tasto
Key06:
	movlw	KB_ESCAPE			; ESC, TAB, +, -
	subwf	TxData1, W
	bz		Key10
	movlw	KB_TAB
	subwf	TxData1, W
	bz		Key10
	movlw	'+'
	subwf	TxData1, W
	bz		Key10
	movlw	'-'
	subwf	TxData1, W
	bnz		KeySnd
Key10:
	movlw	SKPALL				; tasti cursore da tastierino numerico
	andwf	TxFlag, W			; non vengono testati per SHIFT/CTRL/ALT
	sublw	SKPALL
	bz		KeySnd
	movlw	LOW SPCKEY			; imposta ptr tabella tasti speciali 
	movwf	TBLPTRL
	movlw	HIGH SPCKEY
	movwf	TBLPTRH
	setf	E1Cnt				; indice tabella (inizializzato a 0xFF)
Key11:
	incf	E1Cnt, F			; update indice
	tblrd	*+					; legge byte da memoria ed incrementa ptr
	movlw	0xFF				; test fine tabella
	subwf	TABLAT, W			
	bz		KeySnd				; termina sequenza
	ldw		TxData1				; confronta con tasto
	subwf	TABLAT, W
	bnz		Key11				; loop
	movlw	B'11000000'
	andwf	CtlSt, W			; test ALT e/o CTRL
	bz		Key13				; test SHIFT
	sublw	B'11000000'			; CTRL + ALT ?
	bnz		Key12				; NO
	movlw	0x03				; CTRL + ALT
	bra		Key15
Key12:
	movlw	0x01
	btfsc	CtlSt, AALTB
	bra		Key15				; ALT
	movlw	0x02
	bra		Key15				; CTRL
Key13:
	btfss	ShfSt, ASHFB
	bra		KeySnd				; no SHIFT - non modifica tasto
	movlw	0x00
Key15:
	bcf		STATUS, C
	rlcf	E1Cnt, F
	rlcf	E1Cnt, F			; * 4
	addwf	E1Cnt, F	
	movlw	LOW SPCKEY2			; imposta ptr tabella SPCKEY2
	addwf	E1Cnt, W			; aggiunge indice
	movwf	TBLPTRL
	movlw	HIGH SPCKEY2
	movwf	TBLPTRH
	clrf	WREG
	addwfc	TBLPTRH, F
	tblrd	*+					; codice
	ldw		TABLAT
	bz		KeySnd				; tasto null
	movwf	TxData1
	movlw	LOW SPCKEY3			; imposta ptr tabella SPCKEY3
	addwf	E1Cnt, W			; aggiunge indice
	movwf	TBLPTRL
	movlw	HIGH SPCKEY3
	movwf	TBLPTRH
	clrf	WREG
	addwfc	TBLPTRH, F
	tblrd	*+					; codice
	ldw		TABLAT
	movwf	TxData0
	movlw	0xC0
	subwf	TxData0, W
	bz		KbdProc60

KeySnd:
	btfsc	RxFlag, RXBRK		; test tasto rilasciato
	bsf		TxData0, KEYRLS 	; SI - imposta bit rilascio
	btfss	TxData0, KEYTGL		; invia flag toggle low ?
	bra		PutKey				; NO
	ldw		TglSt
	andlw	B'00001111'
	iorwf	TxData0, F			; 4 bit bassi corrispondono ai 4 bit di TglSt
	bra		PutKey
	
KbdProc60:
	; setta NMI all' host
	clrf	Caps				; contatore temporaneo
KbdProc61:
	btfss	PORTA, HLNMI		; host busy
	bra		NMISnd				; NMI non attivo - attiva
	incfsz	Caps, F				; attende host libero
	bra		KbdProc61
	goto	ClrRX				; scarta codice NMI
NMISnd:
	movlw	0x02				; porta NMI
	movwf	LATE				; indirizzo KFR
	ldw		TxData1				; codice NMI
	rcall	KFRWrite			; codice in porta NMI
	bcf		PORTC, KFRSSN		; attiva NMI host
	nop
	bsf		PORTC, KFRSSN
	bra		ClrRX

GetKeyTbl:
	movlw	LOW KEYTBL			; imposta ptr tabella conversione KEYTBL 
	addwf	RxCode, W			; aggiunge indice
	movwf	TBLPTRL
	movlw	HIGH KEYTBL
	movwf	TBLPTRH
	clrf	WREG
	addwfc	TBLPTRH, F
	tblrd	*					; codice tasto
	ldw		TABLAT
	movwf	TxData1				; codice tasto
	movlw	LOW FLGTBL			; imposta ptr tabella conversione FLGTBL 
	addwf	RxCode, W			; aggiunge indice
	movwf	TBLPTRL
	movlw	HIGH FLGTBL
	movwf	TBLPTRH
	clrf	WREG
	addwfc	TBLPTRH, F
	tblrd	*					; flag tasto
	ldw		TABLAT
	movwf	TxData0				; flag tasto
	return


KFRWrite:
	movwf	LATD			; byte da scrivere nel registro
	clrf	DDRD			; porta D in uscita
	nop
	nop						; dato stabile
	bcf		LATC, KFRWE		; impulso write
	nop
	bsf		LATC, KFRWE
	nop
	return

KFRRead:
	setf	DDRD			; Porta D input
	clrf	LATE			; indirizzo file reg. = 0
	movlw	0x04			; contatore registri
	movwf	KFRRdCnt
	lfsr	FSR0, KFRRd0	; puntatore registro 0
KFRRead1:
	bcf		LATC, KFRRD		; lettura registro
	nop
	nop
KFRRead2:
	ldw		PORTD			; legge porta D
	cpfseq	PORTD			; attende dato stabile
	bra		KFRRead2
	bsf		LATC, KFRRD		; linea lettura alta
	movwf	POSTINC0		; memorizza dato registro
	incf	LATE, f			; incrementa indirizzo
	decfsz	KFRRdCnt, f		; loop
	bra		KFRRead1
	clrf	LATE
	bcf		KFRFlag, KFRFULL	; azzera flag richiesta servizio
	movlw	0xE1				; set lock led
	subwf	KFRRd0, W
	bnz		Cmd0
	rcall	SetLed
	clrf	KFRRd3				; errore ?
	btfsc	KbdFlag, KBLEDOK
	setf	KFRRd3
	bra		CmdRsp
Cmd0:
	movlw	0xC1				; start scanning
	subwf	KFRRd0, W
	bnz		Cmd1
	SCLOCKH
	bra		CmdEnd
Cmd1:
	movlw	0xC2				; RESET
	subwf	KFRRd0, W
	bnz		Cmd2
	bcf		INTCON, GIEH		; disabilita INT high priority
	bcf		INTCON, GIEL		; disabilita INT low priority
	bcf		LATC, KBVOFF	; impulso negativo VCC tastiera
	movlw	TMR0P256		; TMR0 stop, prescaler = 256, modo 16 bit
	movwf	T0CON	
	bcf		INTCON, TMR0IF	; azzera flag timeout
	movlw	HIGH(T1500MS)	; imposta timeout 1500ms
	movwf	TMR0H			; caricare prima TMR0H
	movlw	LOW(T1500MS)
	movwf	TMR0L			; carica TMR0L/H
	bsf		T0CON, TMR0ON	; start Timer 0
Wait1:
	btfss	INTCON, TMR0IF
	bra		Wait1			; timeout

	;bsf		LATC, KBVOFF
	bcf		LATC, KFRSRP	; azzera linea DRQ verso host
	nop
	nop
	bsf		LATC, KFRSRP
	reset	
Cmd2:
	bra		CmdEnd
CmdRsp:
	movlw	0x03
	movwf	LATE
	ldw		KFRRd3
	rcall	KFRWrite
	bra		CmdEnd

CmdEnd:	
	bcf		LATC, KFRSRP	; azzera linea DRQ verso host
	nop
	nop
	bsf		LATC, KFRSRP
	return	

SetLed:
	ldw		TglSt
	movwf	TTglSt
	ldw		KFRRd1
	movwf	TglSt
SetLed1:
	ldw		TglSt
	movwf	LedSt
	rrncf	LedSt, F
	rrncf	LedSt, F
	rrncf	LedSt, F
	rrncf	LedSt, F
	rrncf	LedSt, W
	andlw	B'00000111'
	movwf	LedSt
	ldw		LATB				; salva porta B
	movwf	TPortB
	SCLOCKL						; inibisce TX da tastiera
	bcf		INTCON, GIEH		; disabilita INT high priority
	bcf		INTCON, GIEL		; disabilita INT low priority
	call	SendLed
	btfss	TPortB, KCLKO		; disabilita TX da tast. se era disabilitato	
	SCLOCKL
	bcf		INTCON, INT0IF		; resetta flag INT0
	bsf		INTCON, GIEH		; abilita INT high priority
	bsf		INTCON, GIEL		; abilita INT low priority
	bnc		SetLed2				; comando LED riuscito
	ldw		TTglSt				; errore -> ignora tasto
	movwf	TglSt				; reimposta flag toggle precedente
	ldw		TLockSt
	movwf	LockSt
	bcf		KbdFlag, KBLEDOK	; comando LED fallito
	return						; fail
SetLed2:
	bsf		KbdFlag, KBLEDOK
	return

SndProc:						; invia word nella coda TX ad host
	btfsc	PORTA, HLINT		; host busy ?
	return						; si
	ldw		TxTailL				; imposta ptr head al buffer
	movwf	FSR0L
	ldw		TxTailH
	movwf	FSR0H
	ldw		POSTINC0
	movwf	TxData0
	ldw		POSTINC0
	movwf	TxData1
	clrf	LATE				; indirizzo KFR = 0
	ldw		TxData0				; flag tasto
	rcall	KFRWrite
	movlw	0x01				; indirizzo KFR = 1
	movwf	LATE
	ldw		TxData1				; codice tasto
	rcall	KFRWrite
	bcf		PORTC, KFRSSI		; attiva INT host
	nop
	bsf		PORTC, KFRSSI
	decf	TxCnt, F			; aggiorna contatore word nel buffer TX
	ldw		FSR0L				; aggiorna ptr
	movwf	TxTailL
	ldw		FSR0H
	movwf	TxTailH	
	movlw	LOW(TxBufEnd)		; coda circolare - test puntatore
	subwf	TxTailL, W			; TxTailL - LOW(TxBufEnd)
	movlw	HIGH(TxBufEnd)		; TxTailH - HIGH(TxBufEnd)
	subwfb	TxTailH, W			; C = 0 se TxTail < TxBufEnd
	btfss	STATUS, C			; test buffer
	return						; C = 0 -> OK
	movlw	LOW(TxBuffer)		; reimposta puntatore a inizio buffer
	movwf	TxTailL
	movlw	HIGH(TxBuffer)
	movwf	TxTailH
	return

Start:
	movlb	0x00
	call	InitIO
	bcf		LATA, RA5			; bit READY	

Main:
	btfss	LATB, KCLKO			; se TX disabilitato salta
	bra		Main2
	call	KbdProc
	tstfsz	TxCnt				; coda TX vuota
	rcall	SndProc				; invia word codice ad host
Main2:
	btfsc	KFRFlag, KFRFULL	; host richiede servizio ?
	rcall	KFRRead				; si - lettura registro KFR
	bra		Main

	#include romtbl.inc

	end
 